#!/bin/bash -e
#
# prepare-raspbian-sd:bin/prepare-raspbian-sd-download-image.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -gr PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-config-json-api.sh" || exit $?

declare -g ZIP_CHECKED='false'

function usage {
  cat <<EOF
Download Raspberry Pi OS image (or Raspbian) from 'raspberrypi.org' web site.

Usage
  $( basename "$0" ) --help
  $( basename "$0" ) --download [--full] [--unzip] [--arch ARCH]
  $( basename "$0" ) --download --url IMGURL [--expected-img IMG] [--unzip]

Actions
  -h, --help            Display help message and exit.
  -D, --download        Download new image if needed into directory define by
                        RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY:
                            '$( print_variable RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY )' (Current value)
  -T DIR, --download-to DIR
                        Download new image if needed into given directory
  --url IMGURL          Allow to set image file URL

Parameters
  --arch ARCH           Define machine hardware name (optional). If not set, or empty, use "Raspbian"
                        as source, if set use "Raspberry Pi OS" as source). ARCH could be an empty string,
                        armhf or arm64.
                          --arch '' : Download last available version of raspbian
                          --arch armhf: Download last 32 bits version of "Raspberry Pi OS"
                          --arch arm32: Same as 'armhf'
                          --arch arm64: Download last 64 bits version of "Raspberry Pi OS"
                        (Parameter is ignored if --url is used)
  --desktop,            Same as --full (deprecated)
  --expected-img IMG    Expected image filename (without path) - if specified when using --url allow
                        to not download remote file if local file is present.
  --full                Download full version (desktop version) instead of lite version.
                        (Parameter is ignored if --url is used)
  --unzip               Unzip file after downloading

EOF
}

function init_raspios_raspbian_values {
  local -r arch_type="$1"
  local -r os_type="$2" # 'full' or 'lite'

  if [ -z "${arch_type}" ] ; then
    LAST_VERSION_URL="$( build_last_version_url_for_raspbian "${os_type}" )"
  else
    LAST_VERSION_URL="$( build_last_version_url_for_raspberry_pi_os "${arch_type}" "${os_type}" )"
  fi
  LAST_VERSION_FILENAME="$( basename "${LAST_VERSION_URL}" )"
  EXPECTED_ZIP_FILE="${param_download_image_directory}/${LAST_VERSION_FILENAME}"

  # Remove archive extension
  EXPECTED_IMG_FILE="$( sed -E 's,^(.*)(.xz|.zip|.gz),\1,g' <<<"${EXPECTED_ZIP_FILE}" )"
  # Ensure file ending with .img
  EXPECTED_IMG_FILE="${EXPECTED_IMG_FILE%.img}.img"

  if [ -z "${LAST_VERSION_URL}" ] ; then
    echofatal 'Cannot get last version - Check your Internet connection'
    exit 100
  fi

  declare -gr LAST_VERSION_URL
  declare -gr LAST_VERSION_FILENAME
  declare -gr EXPECTED_ZIP_FILE
  declare -gr EXPECTED_IMG_FILE

  cat <<EOF >&2
LAST_VERSION_URL=${LAST_VERSION_URL}
LAST_VERSION_FILENAME=${LAST_VERSION_FILENAME}
EXPECTED_ZIP_FILE=${EXPECTED_ZIP_FILE}
EXPECTED_IMG_FILE=${EXPECTED_IMG_FILE}
EOF
}

function build_last_version_url_for_raspbian {
  local -r os_type="$1"

  local lite=
  if [ "${os_type}" = 'lite' ] ; then
    lite='lite_'
  fi

  local -r url="https://downloads.raspberrypi.org/raspbian_${lite}latest"

  echoinfo "Loading version from: ${url}" >&2

  curl --head "${url}" |
    grep -i '^Location: ' |
    cut -d ' ' -f2- |
    tr '\r' '\n' |
    head -n 1
}

function build_last_version_url_for_raspberry_pi_os {
  local -r arch_type="$1"
  local -r os_type="$2"

  local identifier=
  case "${os_type}" in
    'lite')
      identifier="lite_${arch_type}_latest";
      ;;

    'full')
      identifier="${arch_type}_latest";
      ;;
  esac
  local -r url="https://downloads.raspberrypi.org/raspios_${identifier}"

  echoinfo "Loading version from: ${url}" >&2

  curl --head "${url}" |
    grep -i '^Location: ' |
    cut -d ' ' -f2- |
    tr '\r' '\n' |
    head -n 1
}

function download_if_needed {
  if is_download_needed ; then
    download || return $?
  fi
}

function is_download_needed {
  exit_if_variable_not_define_or_empty param_download_image_directory
  exit_if_directory_not_exists "${param_download_image_directory}"

  if [ -f "${EXPECTED_ZIP_FILE}" ] ; then
    echookay "File '${EXPECTED_ZIP_FILE}' already exist."
    ls -la --color "${EXPECTED_ZIP_FILE}"
    return 1
  else
    echoinfo 'Update required.'
    return 0
  fi
}

function download {
  wget "${LAST_VERSION_URL}" -O "${EXPECTED_ZIP_FILE}" || return $?

  if [ -z "${param_url}" ] ; then
    wget "${LAST_VERSION_URL}.sha256" -O "${EXPECTED_ZIP_FILE}.sha256" || return $?
  fi
}

function unzip_if_needed {
  if is_unzip_needed ; then
    pushd "${param_download_image_directory}" >/dev/null || return $?

    if [ "${EXPECTED_ZIP_FILE%*.zip}.zip" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        unzip "${EXPECTED_ZIP_FILE}"
      ) || return $?
    elif [ "${EXPECTED_ZIP_FILE%*.gz}.gz" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        gzip --decompress --keep "${EXPECTED_ZIP_FILE}"
      ) || return $?
    elif [ "${EXPECTED_ZIP_FILE%*.xz}.xz" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        xz --decompress --keep "${EXPECTED_ZIP_FILE}"
      ) || return $?
    else
      echowarn "Don't kown how to unzip: '${EXPECTED_ZIP_FILE}'" >&2
    fi
    popd >/dev/null || return $?
  fi
}

function is_unzip_needed {
  exit_if_variable_not_define_or_empty param_download_image_directory
  exit_if_directory_not_exists "${param_download_image_directory}"

  exit_if_file_not_exists "${EXPECTED_ZIP_FILE}"

  local target_img="${EXPECTED_IMG_FILE}"

  if [ -f "${target_img}" ] ; then
    echookay "File '${target_img}' already exist."
    ls -la --color "${target_img}"
    return 1
  else
    echoinfo "unzip required -> '${target_img}'"
    return 0
  fi
}

function check_zip_once {
  if [ "${ZIP_CHECKED}" = 'true' ] ; then
    return 0
  fi

  if [ ! -f "${EXPECTED_ZIP_FILE}.sha256" ] ; then
    echofatal "Cannot find '${EXPECTED_ZIP_FILE}.sha256' - Cannot trust zip file."
    return 100
  fi

  echoinfo "Verifying checksum of '${EXPECTED_ZIP_FILE}'" >&2
  local check256=
  check256="$(
    cd "$( dirname "${EXPECTED_ZIP_FILE}" )" || exit $?
    sha256sum "$( basename "${EXPECTED_ZIP_FILE}" )"
    )" || return $?

  local verify256=
  verify256="$( cat "${EXPECTED_ZIP_FILE}.sha256" )" || return $?

  if [ "${check256}" != "${verify256}" ] ; then
    echofatal "Wrong sha256sum !"
    cat <<EOF >&2
COMPUTED: ${check256}
EXPECTED: ${verify256}
EOF
    return 100
  fi

  ZIP_CHECKED='true'
}

function main {
  local action=
  local param=
  local param_arch=             # 'armhf'|'arm64'|'' also 'arm32' alias for 'armhf'
  local param_download_image_directory=
  local param_expected_img=
  local param_os_type='lite'    # 'full'|'lite'
  local param_unzip=
  local param_url=

  exit_if_command_is_missing \
    curl \
    gzip \
    sha256sum \
    unzip \
    wget

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        exit 0
        ;;

      '-D'|'--download')
        action='DOWNLOAD'
        exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY'
        param_download_image_directory="${RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY}"
        ;;

      '-T'|'--download-to')
        action='DOWNLOAD'
        param_download_image_directory="$( test_P "$@" )" || return $?
        shift
        ;;

      '--arch')
        param_arch="$( test_P "$@" )" || return $?
        shift
        ;;

      '--desktop'|'--full')
        param_os_type='full'
        ;;

      '--unzip')
          param_unzip='true'
          ;;

      '--url')
        param_url='true'
        LAST_VERSION_URL="$( test_P "$@" )" || return $?
        shift
        declare -gr LAST_VERSION_URL
        ;;

      '--expected-img')
        param_expected_img="$( test_P "$@" )" || return $?
        shift
        ;;

      *)
        echoerror "Unknown option: '${param}'"
        usage >&2
        exit 1
        ;;
    esac
  done

  if [ -z "${action}" ] ; then
    echofatal 'Action is missing.'
    usage >&2
    exit 1
  fi

  if [ -z "${param_download_image_directory}" ] ; then
    exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY'
    param_download_image_directory="${RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY}"
  fi

  case "${param_arch}" in
    'armhf'|'arm64'|'') ;; # OK
    'arm32')
       param_arch='armhf' # OK, but name need to be fixed
       ;;
    *)
      echofatal "--arch '${param_arch}' is not valid. Nothing, 'armhf' or 'arm64' is expected."
      exit 1
      ;;
  esac

  case "${param_os_type}" in
    'full'|'lite') ;; # OK
    '') param_os_type='lite' ;; # OK

    *)
      echofatal "--os-type '${param_os_type}' is not valid. 'full' or 'lite' is expected."
      exit 1
      ;;
  esac

  if [ -z "${param_url}" ] ; then
    init_raspios_raspbian_values "${param_arch}" "${param_os_type}"
  else
    local basename=

    basename="$( basename "${LAST_VERSION_URL}" )"

    if [ -n "${param_expected_img}" ] ; then
      EXPECTED_IMG_FILE="${param_download_image_directory}/${param_expected_img}"
    else
      EXPECTED_IMG_FILE=
    fi

    declare -gr EXPECTED_ZIP_FILE="${param_download_image_directory}/${basename}"
    declare -gr EXPECTED_IMG_FILE
  fi

  cat <<EOF
LAST_VERSION_URL     :${LAST_VERSION_URL}
LAST_VERSION_FILENAME:${LAST_VERSION_FILENAME:-__UNDEFINED__}
EXPECTED_ZIP_FILE    :${EXPECTED_ZIP_FILE}
EXPECTED_IMG_FILE    :${EXPECTED_IMG_FILE}
EOF

  case "${action}" in
    'DOWNLOAD')
      download_if_needed

      if [ -z "${param_url}" ] ; then
        check_zip_once || return $?
      else
        echowarn "Cannot trust downloaded file (don't how to get hashcode)."
      fi
      ;;

    *)
      echofatal "Action not valid: '${action}'"
      usage >&2
      exit 1
      ;;
  esac

  if [ "${param_unzip}" = 'true' ] ; then
    if [ -z "${param_url}" ] ; then
      check_zip_once || return $?
    else
      echowarn "Cannot trust downloaded file (don't how to get hashcode)."
    fi

    unzip_if_needed || return $?
  fi

  cat <<EOF
LAST_ZIP_FILE:${EXPECTED_ZIP_FILE}
LAST_IMG_FILE:${EXPECTED_IMG_FILE}
EOF
}

prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY "$( xdg-user-dir DOWNLOAD )"

main "$@" || exit $?
